/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#include"infer_old.h"

bool MaxMarginalMode=false;

char *FTypeNames[6]={ "Table", "Gaussian", "MixOfGauss", "UnscentedTransform", "Linear", "NonLinear" };

//===========================================================================
//
// helpers & algorithms
//

void getIndices(uintA& list,const uintA& id,const uintA& dim,const uintA& mid){
  list.clear();
  uint i=0,k=0,d=0,j;
  for(k=0;k<mid.N;k++){
    for(i=0,d=0;i<id.N;i++){ if(id(i)==mid(k)) break; d+=dim(i); }
    CHECK(i<id.N,"requesting marginal variable "<<mid(k) <<" in factor over variables " <<id);
    //while(id(i)<mid(k) && i<id.N){  i++; }
    //if(i==id.N) return;
    for(j=0;j<dim(i);j++){
      list.append(d);
      d++;
    }
  }
  INFER_DEBUG(3,cout <<"** getIndices: id=" <<id <<" mid=" <<mid <<" dim=" <<dim <<" list=" <<list <<endl);
}

void getPick(uintA& pick,const uintA& id,const uintA& mid){
  uint i=0,k=0;
  pick.resize(mid.N);
  for(k=0;k<mid.N;k++){
    for(i=0;i<id.N;i++) if(id(i)==mid(k)) break;
    CHECK(i<id.N,"getPick error: id="<<id <<" mid=" <<mid);
    pick(k)=i;
  }
  INFER_DEBUG(3,cout <<"** getPick: id=" <<id <<" mid=" <<mid <<" pick=" <<pick <<endl);
}


uint passMessage(factorEdge *e,Factor *efrom,Factor *eto,FwdBwd dir){
  CHECK(e->fwd && e->bwd,"no message factors allocated yet!");
  Factor *barf1=0,*barf2=0,*m3=0,*nu=0;
  Factor *mu_fwd,*mu_bwd,*f1,*f2;
  if(dir==FORWARD){
    mu_fwd=e->fwd; mu_bwd=e->bwd; f1=efrom; f2=eto;
  }else{
    mu_fwd=e->bwd; mu_bwd=e->fwd; f1=eto;   f2=efrom;
  }
  INFER_DEBUG(1,cout <<"*** MESSAGE PASSING " <<efrom->id <<"->" <<eto->id <<" (edgeLabel=" <<(char)e->label <<", dir=" <<dir <<")" <<endl;)
  INFER_DEBUG(2,cout <<"original mu_fwd=" <<*mu_fwd <<" mu_bwd=" <<*mu_bwd <<endl;)

  //barf1 = \bar f_1
  f1->marginal(barf1,mu_fwd->id);
  INFER_DEBUG(2,cout <<"bar f_1=" <<*barf1;)

  //nu = \bar f_1 / mu_fwd mu_bwd
  nu = mu_fwd->clone();
  nu->multiply(*mu_bwd);
  INFER_DEBUG(2,cout <<"mu_fwd mu_bwd = " <<*nu <<endl;)
  if(*barf1==*nu){ INFER_DEBUG(1,cout <<"no message passing necessary" <<endl); return 0; }
  nu->invMultiply(*barf1);
  INFER_DEBUG(2,cout <<"nu = " <<*nu <<endl;)

#if 1 //you can do this if there is no approx
  f2->multiply(*nu);  //this should NOT be APPROXIMATE!!!
  mu_fwd->multiply(*nu);
#else //otherwise you need to measure what information has been passed!
  f2->marginal(barf2,mu_fwd->id);
  f2->multiply(*nu);  //this might be APPROXIMATE!!!
  f2->marginal(m3,mu_fwd->id);

  mu_fwd->multiply(*m3);
  mu_fwd->divide(*barf2);
#endif
  delete barf1;
  delete nu;
  if(barf2){
    delete barf2;
    delete m3;
  }

#if 0
  if(dir==FORWARD){
    checkConsistency(eto);
  }else{
    checkConsistency(efrom);
  }
#endif

  INFER_DEBUG(3,MT::wait());

  return 1;
}

Factor *new_factor(FType type,variableA& V,uintA& ids){
  switch(type){
  case TableFT:
    return new TableFac(V,ids);
    break;
  case GaussFT:
    NIY;
    //return new GaussFac(V,ids);
    break;
  case SwitchFT:
    return new SwitchFac(V,ids);
    break;
  default:
    NIY;
  }
  return 0;
}

void allocateEdgeFactors(FactorModel& M){
  uint i;
  uintA id;
  for(i=0;i<M.E.N;i++){
    setSection(id,M.from(i)->id,M.  to(i)->id);
    CHECK(id.N>0,"no intersection between two connected factors");
    if(!M.E(i).fwd){ M.from(i)->marginal(M.E(i).fwd,id); M.E(i).fwd->uniform(); }
    if(!M.E(i).bwd){ M.  to(i)->marginal(M.E(i).bwd,id); M.E(i).bwd->uniform(); }
  }
}

void storeOrgFactors(FactorModel& M){
  uint i;
  for(i=0;i<M.N;i++){
    if(!M.elem(i).org) M.elem(i).org=M.elem(i)->clone();
    else M.elem(i).org->set(*(M.elem(i).f));
  }
}

void checkConsistency(FactorModel& M);

bool checkEquilibrium(FactorModel& M){
  uint i;
  Factor *f0,*f1,*f2;

  for(i=0;i<M.E.N;i++){
    f0 = M.E(i).fwd->clone();  f0->uniform();
    f1 = M.E(i).bwd->clone();  f1->uniform();
    f2 = M.E(i).bwd->clone();  f2->multiply(*(M.E(i).fwd)); //product of M.E(i).fwd*M.E(i).bwd
    M.from(i)->marginal(f0,M.E(i).fwd->id); //marginal of M.E(i).from
    M.from(i)->marginal(f1,M.E(i).fwd->id); //marginal of M.E(i).to
    if(!(*f0==*f2) || !(*f0==*f1)){
      cout <<"factor graph equilibrium check - FAILURE!" <<endl;
      if(!(*f0==*f1)) cout <<*f0 <<"!=" <<*f1 <<endl;
      if(!(*f0==*f2)) cout <<*f0 <<"!=" <<*f2 <<endl;
      return false;
    }
    delete f0;
    delete f1;
    delete f2;
  }
  cout <<"factor graph is in equilibrium - SUCCESS!" <<endl;
  return true;
}

void checkConsistency(FactorModel& M,uint i){
  factorNode *n;
  factorEdge *e;
  Factor *f;
  uint j;
  n = &M.elem(i);

  INFER_DEBUG(3,cout <<"*** checking consistency" <<endl;)
  //check if id's and dim's are in argreement
  uintA list;
  for(j=0;j<M.E.N;j++) if(M.E(j).to==i){
    e=&M.E.elem(j);
    CHECK(e->fwd && e->bwd,"no messages allocated yet");
    CHECK(e->fwd->id==e->bwd->id && e->fwd->dim==e->bwd->dim,"fwd and bwd messages disagree on variables: fwd.id=" <<e->fwd->id <<" bwd.id=" <<e->bwd->id <<" fwd.dim=" <<e->fwd->dim <<" bwd.dim=" <<e->bwd->dim );
    getIndices(list,n->f->id,n->f->dim,e->fwd->id);
    CHECK(list.N>0,"message and factor don't share variables");
  }

  if(n->f->type==LinearFT) return;
  if(n->f->type==SwitchFT) return;

  //check if the factor is equal to the
  //original potential multipled by all incoming messages
  CHECK(n->org,"original factor not stored - cannot check consistency");
  INFER_DEBUG(3,cout <<"original factor=" <<*(n->org) <<endl;)
  f = n->org->clone();
  for(j=0;j<M.E.N;j++) if(M.E(j).to==i){
    e=&M.E.elem(j);
    //for_in_edges(e,n){
    f->multiply(*(e->fwd));
    INFER_DEBUG(3,cout <<"in messages=" <<j <<' ' <<*(e->fwd) <<endl;)
  }
  for(j=0;j<M.E.N;j++) if(M.E(j).fr==i){
    e=&M.E.elem(j);
    //for_out_edges(e,n){
    f->multiply(*(e->bwd));
    INFER_DEBUG(3,cout <<"out messages=" <<j <<' ' <<*(e->bwd) <<endl;)
  }
  INFER_DEBUG(3,cout <<"product of messages with original factor=" <<*f <<endl;)
  INFER_DEBUG(3,cout <<"current=" <<*(n->f) <<endl;)

  CHECK(*n->f == *f,"factor graph consistency check failed: org=:" <<*(n->org) <<" current=" <<*n->f <<" effective=" <<*f);
  delete f;
}

void checkConsistency(FactorModel& M){
  uint i;
  for(i=0;i<M.N;i++) checkConsistency(M,i);
  cout <<"factor graph is consistent     - SUCCESS!" <<endl;
}

void passLabeledEdges(FactorModel& M,int label,FwdBwd dir){
  uint i;
  if(dir==FORWARD){
    for(i=0;i<M.E.N;i++) if(M.E(i).label==label){
      passMessage(&M.E(i),M.from(i),M.to(i),FORWARD);
      //INFER_DEBUG(1,cout <<"passing " <<i <<(char)label <<'(' <<M.from(i).name <<"->" <<M.from(i).name <<')' <<endl);
    }
  }else{
    for(i=M.E.N;i--;) if(M.E(i).label==label){
      passMessage(&M.E(i),M.from(i),M.to(i),BACKWARD);
      //INFER_DEBUG(1,cout <<"passing " <<i <<(char)label <<'(' <<M.to(i).name <<"->" <<M.to(i).name <<')' <<endl);
    }
  }
}

void unrollDBN(FactorModel& M,uint T,uint dV){
  FactorModel A;
  A=M;
  M.resize(T,A.N);
  M.E.clear();
  M.V.resize(T,dV);

  M.T=T;
  M.dF=A.N;
  M.dV=dV;

  uint t,i,j=0;
  for(t=0;t<T;t++){
    for(i=0;i<A.N;i++){
      M(t,i) = A(i)->clone();
      M(t,i)->id += t*dV;
      //M(t,i).name <<'.' <<t;
    }
    for(i=0;i<dV;i++){
      M.V(t,i) = A.V(i);
      M.V(t,i).name <<'.' <<t;
    }
    for(i=0;i<A.E.N;i++){
      if(!A.E(i).temporal){
        M.E.resizeCopy(j+1);
        M.E(j)=A.E(i);
        M.E(j).fr += t*A.N;
        M.E(j).to += t*A.N;
        if(M.E(j).fwd){ M.E(j).fwd->id += t*dV; }
        if(M.E(j).bwd){ M.E(j).bwd->id += t*dV; }
        j++;
      }else if(t<T-1){
        M.E.resizeCopy(j+1);
        M.E(j)=A.E(i);
        M.E(j).fr +=  t    *A.N;
        M.E(j).to += (t+1) *A.N;
        if(M.E(j).fwd){ M.E(j).fwd->id += t*dV; }
        if(M.E(j).bwd){ M.E(j).bwd->id += t*dV; }
        j++;
      }
    }
  }
}

//===========================================================================
//
// Factor instantiations -- Table
//

double robustDevide(double a,double b){
  if(!a) if(!b) return 1.; else return 0.;
  //if(a==b) return 1.;
  if(b) return a/b;
  HALT("division by zero: " <<a <<"/" <<b);
}
	
TableFac::TableFac(){
  type = TableFT;
}

/*TableFac::TableFac(uint _var,uint _dim){
  NIY;
  type = TableFT;
  id.resize(1);  id=_var;
  dim.resize(1); dim=_dim;
  P.tensorResize(dim);
  uniform();
}

TableFac::TableFac(uint _var,uint _dim,arr &q){
  NIY;
  type = TableFT;
  id.resize(1);  id=_var;
  dim.resize(1); dim=_dim;
  set(q);
}*/

TableFac::TableFac(const uintA &_var,const uintA &_dim){
  setBase(TableFT,_var,_dim);
  P.tensorResize(dim);
  uniform();
}

TableFac::TableFac(const uintA &_var,const uintA &_dim,const arr &q){
  setBase(TableFT,_var,_dim);
  set(q);
}

TableFac::TableFac(const variableA &V,const uintA &_var){
  type = TableFT;
  id=_var;
  dim.resize(id.N);
  for(uint i=0;i<id.N;i++) dim(i)=V.elem(id(i)).d;
  P.tensorResize(dim);
  uniform();
}

TableFac::TableFac(const variableA &V,const uintA &_var,const arr &q){
  type = TableFT;
  id=_var;
  dim.resize(id.N);
  for(uint i=0;i<id.N;i++) dim(i)=V.elem(id(i)).d;
  set(q);
}

Factor *TableFac::clone(){
  TableFac *f=new TableFac;
  f->set(*this);
  return f;
}

void TableFac::uniform(){
  P=1.;
  logP=0.;
  lognormScale(P,logP);
}

void TableFac::random(){
  P.tensorResize(dim);
  rndUni(P,.1,1.,false);
  logP=rnd.uni();
  lognormScale(P,logP);
}

#if 0
void TableFac::multiply(const Factor& q){
  CHECK(q.type==TableFT,"wrong factor type");
  CHECK(q.id==id,"wrong variables: factor-ids=" <<id <<" multiplier-ids=" <<q.id);
  mult(P,P,((const TableFac&)q).P);
  logP += ((const TableFac&)q).logP;
  lognormScale(P,logP);
}
#endif

void TableFac::set(const Factor& _q){
  CHECK(_q.type==TableFT,"wrong factor type");
  const TableFac& q=(const TableFac&)_q;
  id=q.id;
  dim=q.dim;
  P = q.P;
  logP = q.logP;
  lognormScale(P,logP);
}

void TableFac::set(const arr& p){
  P = p;
  CHECK(P.N==product(dim) && dim==P.d,"TableFac set with ill-dimensioned array");
  logP = 0.;
  lognormScale(P,logP);
}

void TableFac::invMultiply(const Factor& _q){
  CHECK(_q.type==TableFT,"wrong factor type");
  const TableFac& q=(const TableFac&)_q;
  for(uint i=0;i<P.N;i++) P.elem(i)=robustDevide(q.P.elem(i),P.elem(i));
  logP = -logP + q.logP;
  lognormScale(P,logP);
}

void TableFac::divide(const Factor& _q){
  CHECK(_q.type==TableFT,"wrong factor type");
  const TableFac& q=(const TableFac&)_q;
  for(uint i=0;i<P.N;i++) P(i)=robustDevide(P(i),q.P(i));
  logP -= q.logP;
  lognormScale(P,logP);
}

void TableFac::weightedAdd(double w,const Factor& _q){
  CHECK(_q.type==TableFT,"wrong factor type");
  CHECK(_q.dim==dim,"wrong dimensions?")
  const TableFac &q=((const TableFac&)_q);
  w *= ::exp(q.logP-logP);
  for(uint i=0;i<P.N;i++){
    P.elem(i) = P.elem(i) + w*q.P.elem(i);
  }
  lognormScale(P,logP);
}

void TableFac::multiply(const Factor& _m){
  CHECK(_m.type==TableFT,"wrong factor type");
  const TableFac& m=(const TableFac&)_m;

  if(m.id==id){
    mult(P,P,m.P);
  }else{
    uintA pick;
    getPick(pick,id,m.id);
    tensorMultiply(P,m.P,pick);
  }
  logP += m.logP;
  lognormScale(P,logP);
}


void TableFac::evidence(uint e){
  P=0.;
  P(e)=1.;
  logP=0.;
}

void TableFac::evidence(const arr& e){
  NIY;
}

void TableFac::marginal(Factor*& m,uintA& mid){
  if(!m){
    m=new TableFac;
    m->id=mid;
  }
  if(!m->id.N) m->id=mid;
  marginal(*m);
}

void TableFac::marginal(Factor &_m){
  CHECK(_m.type==TableFT,"wrong factor type");
  TableFac& m=(TableFac&)_m;
  uintA pick;
  getPick(pick,id,m.id);
  m.dim.resize(m.id.N);
  for(uint k=0;k<m.id.N;k++) m.dim(k)=dim(pick(k));
  if(!MaxMarginalMode){
    tensorMarginal(m.P,P,pick);
  }else{
    tensorMaxMarginal(m.P,P,pick);
    //do we need to normalize in the max-marginal case?
    //normalizeDist(Y);
  }
  m.logP=logP;
  lognormScale(m.P,m.logP);
}

bool TableFac::operator==(Factor& _q){
  CHECK(_q.type==TableFT,"wrong factor type");
  TableFac& q=(TableFac&)_q;
  //uint i=P.maxIndex();
  //double ratio=P(i)/q.P(i);
  lognormScale(P,logP,true);
  lognormScale(q.P,q.logP,true);
  bool r =
    id==q.id
    && dim==q.dim
    && maxDiff(P,q.P,NULL)<1e-10
    && fabs(logP-q.logP)/(1+fabs(logP)+fabs(q.logP))<1e-10;
  if(!r){
    //MT_MSG("unequal table factors: " <<id <<q.id <<dim <<q.dim <<maxDiff(P,q.P) <<' ' <<logP <<' ' <<q.logP);
  }
  return r;
}

void TableFac::write(std::ostream& os) const{
  //os <<"F" <<type <<" V" <<id <<':' <<dim;
  if(P.N<1000)
  os <<" P" <<P;
  os <<'*' <<::exp(logP) <<"=exp(" <<logP <<")" <<endl;
}


//===========================================================================
//
// Factor instantiations -- Switch
//


SwitchFac::SwitchFac(){
  type = SwitchFT;
}

SwitchFac::SwitchFac(const variableA &V,const uintA &_var){
  type = SwitchFT;
  id=_var;
  dim.resize(id.N);
  uint i;
  for(i=0;i<id.N;i++) dim(i)=V.elem(id(i)).d;
  n=id.N-2;
  d=dim(1);
  CHECK(dim(0)==n,"switching variable dimension != number of variables");
  for(i=0;i<n;i++) CHECK(dim(2+i)==dim(1),"switch: variables have different dimensions");
  Pi.resize(n);
  P.resize(n+1,d);
  Z=0.;
  uniform();
}

SwitchFac::SwitchFac(const variableA &V,const uintA &_var,const arr &q){
  type = SwitchFT;
  id=_var;
  dim.resize(id.N);
  uint i;
  for(i=0;i<id.N;i++) dim(i)=V.elem(id(i)).d;
  n=id.N-2;
  d=dim(1);
  CHECK(dim(0)==n,"switching variable dimension != number of variables");
  for(i=0;i<n;i++) CHECK(dim(2+i)==dim(1),"switch: variables have different dimensions");
  Pi.resize(n); Pi=1.;
  set(q);
}

Factor *SwitchFac::clone(){
  SwitchFac *f=new SwitchFac;
  f->set(*this);
  return f;
}

void SwitchFac::uniform(){
  Pi=1.;
  P=1.;
  Z=0.;
  for(uint i=0;i<n+1;i++) lognormScale(P[i](),Z);
}

void SwitchFac::random(){
  n=id.N-2;
  d=dim(1);
  Pi.resize(n);     rndUni(Pi,.1,1.,false);
  P.resize(n+1,d);  rndUni(P,.1,1.,false);
  Z=rnd.uni();
  lognormScale(Pi,Z);
  for(uint i=0;i<n+1;i++) lognormScale(P[i](),Z);
}

void SwitchFac::set(const Factor& q){
  CHECK(q.type==SwitchFT,"wrong factor type");
  const SwitchFac &f=((const SwitchFac&)q);
  id   = f.id;
  dim  = f.dim;
  n    = f.n;
  d    = f.d;
  Pi   = f.Pi;
  P    = f.P;
  Z    = f.Z;
  for(uint i=0;i<n+1;i++) lognormScale(P[i](),Z);
}

void SwitchFac::set(const arr& p){
  CHECK(p.nd==2,"");
  n = p.d0-1;
  d = p.d1;
  P = p;
  P.reshape(n+1,d);
  Z = 0.;
  for(uint i=0;i<n+1;i++) lognormScale(P[i](),Z);
}

#if 0
void SwitchFac::invMultiply(const Factor& q){
  CHECK(q.type==SwitchFT,"wrong factor type");
  const arr& qP=((const SwitchFac&)q).P;
  for(uint i=0;i<P.N;i++) if(qP.elem(i)) P.elem(i)=qP.elem(i)/P.elem(i); else P.elem(i)=0.;
  logP = -logP + ((const SwitchFac&)q).logP;
  lognormScale(P,logP);
}

void SwitchFac::divide(const Factor& q){
  CHECK(q.type==SwitchFT,"wrong factor type");
  const arr& qP=((const SwitchFac&)q).P;
  for(uint i=0;i<P.N;i++) if(P(i)) P(i)/=qP(i);
  logP -= ((const SwitchFac&)q).logP;
  lognormScale(P,logP);
}

void SwitchFac::weightedAdd(double w,Factor& q){
  CHECK(q.type==SwitchFT,"wrong factor type");
  CHECK(q.dim==dim,"wrong dimensions?")
  SwitchFac &f=((SwitchFac&)q);
  w *= ::exp(f.logP-logP);
  for(uint i=0;i<P.N;i++){
    P.elem(i) = P.elem(i) + w*f.P.elem(i);
  }
  lognormScale(P,logP);
}
#endif

void SwitchFac::multiply(const Factor& _m){
  CHECK(_m.type==TableFT,"wrong factor type");
  const TableFac& m=(const TableFac&)_m;

  uintA pick;
  getPick(pick,id,m.id);
  CHECK(pick.N==1,"can only multiply single marginals on switch factors so far");
  uint i=pick(0);
  if(i==0){
    mult(Pi,Pi,m.P);
    Z += m.logP;
    lognormScale(Pi,Z);
  }else{
    i--;
    mult(P[i](),P[i],m.P);
    Z += m.logP;
    lognormScale(P[i](),Z);
  }
}

void SwitchFac::marginal(Factor*& m,uintA& mid){
  if(!m){
    m=new TableFac;
    m->id=mid;
  }
  if(!m->id.N) m->id=mid;
  marginal(*m);
}

void SwitchFac::marginal(Factor &_m){
  CHECK(_m.type==TableFT,"wrong factor type");
  TableFac& m=(TableFac&)_m;

  uintA pick;
  getPick(pick,id,m.id);
  CHECK(pick.N==1,"can only multiply single marginals on switch factors so far");

  m.dim.resize(pick.N);
  for(uint k=0;k<m.id.N;k++) m.dim(k)=dim(pick(k));

  uint i,j;
  arr pi(n),px(d);

  if(pick(0)==0){       //marginal P(i)
    for(i=0;i<n;i++) pi(i) = scalarProduct(P[0],P[1+i]);
    pi *= Pi;
    m.P=pi; m.logP=Z;
    lognormScale(m.P,m.logP);
  }else if(pick(0)==1){ //marginal P(X)
    px.setZero();
    for(i=0;i<n;i++) px += Pi(i) * P[1+i];
    px *= P[0];
    m.P=px; m.logP=Z;
    lognormScale(m.P,m.logP);
  }else{                //marginal P(X_i)
    j=pick(0)-2;
    double b=0;
    for(i=0;i<n;i++) pi(i) = scalarProduct(P[0],P[1+i]);
    for(i=0;i<n;i++) if(i!=j) b += Pi(i) * pi(i);
    px = Pi(j) * P[0];
    px += b; //add a scalar to the vector...
    px *= P[1+j];
    m.P=px; m.logP=Z;
    lognormScale(m.P,m.logP);
  }
}

bool SwitchFac::operator==(Factor& _q){
  CHECK(_q.type==SwitchFT,"wrong factor type");
  SwitchFac& q=(SwitchFac&)_q;
  //uint i=P.maxIndex();
  //double ratio=P(i)/q.P(i);
  //lognormScale(P,logP,true);
  //lognormScale(q.P,q.logP,true);
  bool r =
    id==q.id
    && dim==q.dim
    && maxDiff(Pi,q.Pi,NULL)<1e-10
    && maxDiff(P,q.P,NULL)<1e-10
    && fabs(Z-q.Z)/(1+fabs(Z)+fabs(q.Z))<1e-10;
  if(!r){
    MT_MSG("unequal table factors: " <<id <<q.id <<dim <<q.dim <<maxDiff(P,q.P,NULL) <<' ' <<Z <<' ' <<q.Z);
  }
  return r;
}

void SwitchFac::write(std::ostream& os) const{
  //os <<"F" <<type <<" V" <<id <<':' <<dim;
  if(P.N<1000)
  os <<" P" <<P;
  os <<'{' <<Z <<'}' <<endl;
}


