#include "mdp.h"
#include "mstep.h"
#include "infer_old.h"

#define DEBUGEM(x) //x
#ifndef rescaleRewards
#  define rescaleRewards false
#endif

void glDisplayGrey   (const arr &x,uint d0,uint d1,bool wait,uint win);
void glDisplayRedBlue(const arr &x,uint d0,uint d1,bool wait,uint win);

#define Pt(t) ::pow(mdp.gamma,t)

double EMstep_lev2_align_struct2(
      const MDP& mdp,
      FSC_lev2& fsc,
      uint T,
      bool stepDebug,
      bool maxMstep,
      bool adaptP0){
  
  CHECK(mdp.Px.nd==1 && mdp.Pxax.nd==3 &&
        mdp.Pyxa.nd==3 && mdp.Rax.nd==2 &&
        fsc.Pa0.nd==2 && fsc.P01y0.nd==4,"");
  
  uint
    dx=mdp.Pxax.d0,
    da=mdp.Pxax.d1,
    dy=mdp.Pyxa.d0,
    d0=fsc.P0.N,
    d1=fsc.P1.N;

  uint t;

  MaxMarginalMode=false;

  arr mdp_Rax = mdp.Rax;
  double Rmin=mdp_Rax.min(),Rmax=mdp_Rax.max();
  if(rescaleRewards || (!maxMstep && Rmin<0.)){
    if(!rescaleRewards) MT_MSG("can't handle neg rewards in case of exact M-step -- I'm enforcing rescaling of rewards!");
    for(uint i=0;i<mdp_Rax.N;i++) mdp_Rax.elem(i) = (mdp_Rax.elem(i)-Rmin)/(Rmax-Rmin);
  }else{
    Rmin=0.; Rmax=1.;
  }

  //step 0: eliminate a
  arr Px0x(dx,d0,dx), R0x(d0,dx), Pyx0(dy,dx,d0);
  tensor(Px0x, mdp.Pxax, TUP(0,3,2), fsc.Pa0, TUP(3,1), 1);
  tensor(Pyx0, mdp.Pyxa, TUP(0,1,3), fsc.Pa0, TUP(3,2), 1);
  tensor(R0x , mdp_Rax , TUP(2,1)  , fsc.Pa0, TUP(2,0), 1);

  //step 1: eliminate y:
  arr P01x0,P1x01;
  P01x0.tensorResize(TUP(d0,d1,dx,d0));
  P1x01.tensorResize(TUP(d1,dx,d0,d1));
  tensor(P01x0, fsc.P01y0, TUP(0,1,4,3), Pyx0, TUP(4,2,3), 1);
  tensor(P1x01, fsc.P1y01, TUP(0,4,2,3), Pyx0, TUP(4,1,2), 1);

  DEBUGEM( Px0x.tensorCheckCondNormalization(1);
	   Pyx0.tensorCheckCondNormalization(1); 
	   P01x0.tensorCheckCondNormalization(1);
	   P1x01.tensorCheckCondNormalization(1); );
	
  //step 2: create 2-time-slice factor graph
  uint x,y,n0,n1,a,x_,y_,n0_,n1_,a_; //id's for the variables
  FactorModel M;
  //variables
  x   = M.variable(dx ,"state(t)");
  y   = M.variable(dy ,"observation(t)");
  n1  = M.variable(d1 ,"node1(t)");
  n0  = M.variable(d0 ,"node0(t)");
  a   = M.variable(da ,"action(t)");
  x_  = M.variable(dx ,"state(t+1)");
  y_  = M.variable(dy ,"observation(t+1)");
  n1_ = M.variable(d1 ,"node1(t+1)");
  n0_ = M.variable(d0 ,"node0(t+1)");
  a_  = M.variable(da ,"action(t+1)");
  //factors: two
  M.factor(TableFT,TUP(x_,n0,n1,x));
  M.factor(TableFT,TUP(n1_,x_,n0,n1));
  M.factor(TableFT,TUP(n0_,n1_,x_,n0));
  
  M(0)->multiply(TableFac(M.V,TUP(        x_,n0,x)  ,Px0x));
  M(1)->multiply(TableFac(M.V,TUP(    n1_,x_,n0,n1) ,P1x01));
  M(2)->multiply(TableFac(M.V,TUP(n0_,n1_,x_,n0)    ,P01x0));
  
  ((TableFac*)(M(1).f))->P.tensorResize(TUP(d1,dx,d0,d1));
  ((TableFac*)(M(2).f))->P.tensorResize(TUP(d0,d1,dx,d0));
  //edges
  M.arc(0,1); //(all edges implicitly labeled `0')
  M.arc(1,2); //(all edges implicitly labeled `0')
  //M.arcs(TUP(1,2)); //(all edges implicitly labeled `0')
  allocateEdgeFactors(M);
  //storeOrgFactors(M);

  //TableFac tmp(M.V,TUP(x));

  //step 3: make 2 copys of the factor graph for fwd and bwd iteration...
  FactorModel Mfwd,Mbwd;
  Mbwd=Mfwd=M;

  //step 4: initialize fwd and bwd models:
  Mfwd(0)->multiply(TableFac(M.V,TUP(x )  ,mdp.Px));
  Mfwd(0)->multiply(TableFac(M.V,TUP(n1)  ,fsc.P1));
  Mfwd(0)->multiply(TableFac(M.V,TUP(n0)  ,fsc.P0));
  Mbwd(0)->multiply(TableFac(M.V,TUP(n0,x),R0x));
  DEBUGEM(
    storeOrgFactors(Mfwd);
    storeOrgFactors(Mbwd);
    checkConsistency(Mfwd);
    checkConsistency(Mbwd); );

  //-- prepare some buffers for fwd-bwd
  //buffers for closing the loop
  TableFac head  (M.V,TUP(n0 ,n1, x ));
  TableFac tail  (M.V,TUP(n0 ,n1, x ));
  Mfwd(0)->marginal(head);
  Mbwd(0)->marginal(tail);
  
  //overlab measures
  arr L(2*T); L.setZero();
  L(0) = sum(head.P % tail.P)*::exp(head.logP+tail.logP);
  arr La(T),Lb(T); La.setZero(); Lb.setZero();

  //collectors
  TableFac alpha(M.V,TUP(n0 ,n1, x )); alpha.P=0.; alpha.logP=0.;
  Mfwd(0)->marginal(alpha);
  TableFac beta (M.V,TUP(n0_,n1_,x_)); beta .P=0.; beta .logP=0.;
  //Mbwd(1)->marginal(beta);

  //marginals for display
  TableFac xmarg(M.V,TUP(x)),nmarg(M.V,TUP(n0));


  //***********!!!!!!!!!!!********** switch to max product?
  MaxMarginalMode=false;

  //step 5: iterate
  MT::resetTimer();
  for(t=0;t<T;t++){
    //display some marginal
    if(stepDebug){
      //display head's x-marginal
			cout <<"t=" <<t;
      xmarg.id=x;
      head.marginal(xmarg);
      glDisplayGrey(xmarg.P*::exp(xmarg.logP),global_maze.d0,global_maze.d1,false,0);
			cout <<" alpha_x_norm=" <<::exp(xmarg.logP);
      La(t) = xmarg.P(xmarg.P.N-1)*::exp(xmarg.logP);
      //display head's n-marginal
      nmarg.id=n0;
			head.marginal(nmarg);
      glDisplayGrey(nmarg.P,1,d0,false,2);
			cout <<" alpha_n_norm=" <<::exp(nmarg.logP);
      //display tail's x-marginal
      if(t>0) xmarg.id=x_; else xmarg.id=x;
      tail.marginal(xmarg);
      glDisplayRedBlue(xmarg.P*::exp(xmarg.logP),global_maze.d0,global_maze.d1,false,1);
			cout <<" beta_x_norm=" <<::exp(xmarg.logP);
      Lb(t) = xmarg.P(0)*::exp(xmarg.logP);
			cout <<endl;
			MT::wait();
    }

    //propagate fwd and bwd
#if 1 //expensive way
    passLabeledEdges(Mfwd,0,FORWARD);
    passLabeledEdges(Mbwd,0,BACKWARD);
#else //maybe cheaper way
    Mfwd(0)->marginal(nu);
    Mfwd(1)->multiply(nu);
    Mbwd(1)->marginal(nu);
    Mbwd(0)->multiply(nu);
#endif
    DEBUGEM( passLabeledEdges(Mbwd,0,FORWARD); //don't need this really...
	     checkConsistency(Mfwd);
	     checkEquilibrium(Mfwd);
	     checkConsistency(Mbwd);
	     checkEquilibrium(Mbwd); );

    //get tail from backward model
    tail.id = TUP(n0 ,n1, x );
    Mbwd(0)->marginal(tail);

    //likelihood for T=t+(t-1), before updating head
    L(t+t)   = sum(head.P % tail.P)*::exp(head.logP+tail.logP);

    //get head from the forward model
    head.id = TUP(n0_,n1_,x_);
    Mfwd(2)->marginal(head);

    //likelihood for T=t+t
    L(t+t+1) = sum(head.P % tail.P)*::exp(head.logP+tail.logP);
    DEBUGEM( cout <<"head log=" <<head.logP <<" tail log=" <<tail.logP <<endl;  );

    //accumulation of alpha and beta
    head.id = TUP(n0 ,n1 ,x );
    tail.id = TUP(n0_,n1_,x_);
    alpha.weightedAdd(Pt(t+1),head);
    beta .weightedAdd(Pt(t+1),tail);

    //reset fwd and bwd models for next iteration
    Mfwd=M;
    Mbwd=M;

    //set head and tail
    Mfwd(0)->multiply(head);
    Mbwd(2)->multiply(tail);
    DEBUGEM( storeOrgFactors(Mfwd);
	     storeOrgFactors(Mbwd);
	     checkConsistency(Mfwd);
	     checkConsistency(Mbwd); );

  }
  cout <<"E: " <<MT::getTimer() <<"sec, M: " <<std::flush;

  //***********!!!!!!!!!!!********** switch to max product
  //MaxMarginalMode=true;

  //***** M-STEP, part 1: compute the expectations
  
  TableFac X1y01(M.V,TUP(    n1_,y_,n0,n1));
  TableFac X01y0(M.V,TUP(n0_,n1_,y_,n0));
  TableFac Xa0  (M.V,TUP(a,n0));
  TableFac X01  (M.V,TUP(n0,n1));
  TableFac X0   (M.V,TUP(n0));

  double logP=0.;
  //double err;
  
  arr A,B,C,D,E,F,G,H;
  
  //(follows exactly the equations in the manuscript)
  //tensor A
  C.resize(dx,da,d0,d1);
  tensor(C, alpha.P, TUP(2,3,4), mdp.Pxax, TUP(0,1,4), 1);
  tensorMultiply(C, fsc.Pa0, TUP(1,2));
  D.resize(dy, dx, d0, d1);
  tensor(D, mdp.Pyxa, TUP(0,1,4), C, TUP(1,4,2,3), 1);
  A.tensorResize(TUP(d0,d1,dy,d0,d1));
  tensor(A, beta.P, TUP(0,1,5), D, TUP(2,5,3,4), 1);
  A *= ::exp(beta.logP);
  
  //tensor B
  E.resize(dy,da,dx);
  tensor(E, mdp.Pxax, TUP(3,1,2), mdp.Pyxa, TUP(0,3,1), 1);
  tensorMultiply(E, mdp_Rax, TUP(1,2));
  F.resize(dy,da,d0,d1);
  tensor(F, alpha.P, TUP(2,3,4), E, TUP(0,1,4), 1);
  B.resize(dy,d0,d1);
  tensor(B, fsc.Pa0, TUP(3,1), F, TUP(0,3,1,2), 1);
  
  //tensor X1
  tensorAdd(A, B, TUP(2,3,4));
  G.resize(d1,dy,d0,d1);
  tensor(G, fsc.P01y0, TUP(4,0,1,2), A, TUP(4,0,1,2,3), 1);
  G *= fsc.P1y01;
  lognormScale(G,logP);
  //err=maxDiff(X1y01.P,G);  cout <<"error=" <<err <<endl;  CHECK(err<1.e-4,"");
  X1y01.P = G;  X1y01.logP=logP;
  
  //tensor X0
  G.resize(d0,d1,dy,d0);
  tensor(G, fsc.P1y01, TUP(1,2,3,4), A, TUP(0,1,2,3,4), 1);
  G *= fsc.P01y0;
  lognormScale(G,logP);
  //err=maxDiff(X01y0.P,G);  cout <<"error=" <<err <<endl;  CHECK(err<1.e-4,"");
  X01y0.P = G;  X01y0.logP=logP;
  
  //tensor Xa
  C.resize(d1,dy,dx,d0);
  tensor(C, fsc.P01y0, TUP(4,0,1,3), beta.P, TUP(4,0,2), 1);
  D.resize(dy,dx,d0,d1);
  tensor(D, fsc.P1y01, TUP(4,0,2,3), C, TUP(4,0,1,2), 1);
  E.resize(dy,dx,d0,dx);
  tensor(E, alpha.P, TUP(2,4,3), D, TUP(0,1,2,4), 1);
  F.resize(dx,da,d0,dx);
  tensor(F, mdp.Pyxa, TUP(4,0,1), E, TUP(4,0,2,3), 1);
  G.resize(da,d0);
  tensor(G, mdp.Pxax, TUP(2,0,3), F, TUP(2,0,1,3), 2);
  G *= fsc.Pa0;
  G *= ::exp(beta.logP);
  //second part
  C.resize(d0,dx);
  tensorMarginal(C, alpha.P, TUP(0,2));
  D.resize(da,d0);
  tensor(D, C, TUP(1,2), mdp_Rax, TUP(0,2), 1);
  D *= fsc.Pa0;
  G += D;
  lognormScale(G,logP);
  //err=maxDiff(Xa0.P,G);  cout <<"error=" <<err <<endl;  CHECK(err<1.e-4,"");
  Xa0.P = G;  Xa0.logP=logP;
  
  //tensor X01x (for time step 0!)
  C.resize(d1,dy,dx,d0);
  tensor(C, fsc.P01y0, TUP(4,0,1,3), beta.P, TUP(4,0,2), 1);
  D.resize(dy,dx,d0,d1);
  tensor(D, fsc.P1y01, TUP(4,0,2,3), C, TUP(4,0,1,2), 1);
  E.resize(dx,d0,d1,da);
  tensor(E, mdp.Pyxa, TUP(4,0,3), D, TUP(4,0,1,2), 1);
  F.resize(d0,d1,da,dx);
  tensor(F, mdp.Pxax, TUP(4,2,3), E, TUP(4,0,1,2), 1);
  F *= ::exp(beta.logP);
  //second part
  G.resize(d0,d1,da,dx);
  G=1.;
  tensorMultiply(G,mdp_Rax, TUP(2,3));
  F += G;
  //continue
  G.resize(d0,d1,dx);
  tensor(G, fsc.Pa0, TUP(3,0), F, TUP(0,1,3,2), 1);
  H.resize(d0,d1);
  tensor(H, mdp.Px, TUP(2), G, TUP(0,1,2), 1);
  tensorMultiply(G, fsc.P0, TUP(0));
  tensorMultiply(G, fsc.P1, TUP(1));
  lognormScale(G,logP);
  X01.P = G;  X01.logP=logP;
  tensorMarginal(X0.P, X01.P, TUP(0));
  lognormScale(X0.P, X0.logP);

  
  //***** M-STEP, part 2: update the parameters given the expectations
    
  if(fsc.hierarchical){
    arr Xe01y0, Xe1y01;
    Xe01y0.tensorResize(TUP(2,d0,d1,dy,d0));
    Xe1y01.tensorResize(TUP(2,d1,dy,d0,d1));
    
    tensor(Xe01y0, X01y0.P, TUP(1,2,3,4), fsc.Pe0, TUP(0,4));
    tensor(Xe1y01, X1y01.P, TUP(1,2,3,4), fsc.Pe0, TUP(0,3));
    
    arr XE_01, X_E_0y0, XE_1y1, X_E_11;
    tensorMarginal(XE_01  , Xe01y0[1], TUP(0,1));
    tensorMarginal(X_E_0y0, Xe01y0[0], TUP(0,2,3));
    tensorMarginal(XE_1y1 , Xe1y01[1], TUP(0,1,3));
    tensorMarginal(X_E_11 , Xe1y01[0], TUP(0,3));
    
    if(maxMstep){
      noisyMaxMstep(fsc.PE_01  , XE_01  , 1);
      noisyMaxMstep(fsc.P_E_0y0, X_E_0y0, 1);
      noisyMaxMstep(fsc.PE_1y1 , XE_1y1 , 1);
      noisyMaxMstep(fsc.Pa0    , Xa0.P  , 1);
      if(adaptP0) noisyMaxMstep(fsc.P0     , X0.P   , 1);
    }else{
      standardMstep(fsc.PE_01  , XE_01  , 1);
      standardMstep(fsc.P_E_0y0, X_E_0y0, 1);
      standardMstep(fsc.PE_1y1 , XE_1y1 , 1);
      standardMstep(fsc.Pa0    , Xa0.P  , 1);
      if(adaptP0) standardMstep(fsc.P0     , X0.P   , 1);
    }
      
    hierarchicalToFactors(fsc);
  
  }else{ //(fsc.hierarchical)
    if(maxMstep){
      noisyMaxMstep(fsc.P1y01, X1y01.P, 1);
      noisyMaxMstep(fsc.P01y0, X01y0.P, 1);
      noisyMaxMstep(fsc.Pa0  , Xa0.P  , 1);
      if(adaptP0) noisyMaxMstep(fsc.P0   , X0.P   , 1);
    }else{
      standardMstep(fsc.P1y01, X1y01.P, 1);
      standardMstep(fsc.P01y0, X01y0.P, 1);
      standardMstep(fsc.Pa0  , Xa0.P  , 1);
      if(adaptP0) standardMstep(fsc.P0   , X0.P   , 1);
    }
  }

  //----- rest is cosmetics
  //report
  cout <<MT::getTimer() <<"sec," <<std::flush;
  double Ltot=0.,PtNorm=0.,Texp=0.;
  arr PtL(2*T);
  for(t=0;t<2*T;t++){ PtL(t)=Pt(t)*L(t); Ltot+=PtL(t); Texp+=t*PtL(t); PtNorm+=Pt(t); }
  PtL  /= PtNorm;
  Texp /= Ltot;
  Ltot /= PtNorm; //(in case Pt was not normalized)
  cout <<" P(r=1)=" <<Ltot
       <<", Exp(T)=" <<Texp <<"/"<<::log(Ltot*PtNorm)/::log(mdp.gamma)
       <<", Exp(R)=" <<(Ltot*(Rmax-Rmin)+Rmin)*PtNorm
       <<endl;

  //displays
  if(stepDebug){
    glDisplayGrey(fsc.P01y0,d0*d1*dy,d0,false,2);
    glDisplayGrey(fsc.P1y01,d1*dy*d0,d1,false,3);
    glDisplayGrey(fsc.Pa0 ,da,d0,false,1);
    L.resizeCopy(T); //make shorter for storage...
    write(L,PtL,La,Lb,"z.L");
    //gnuplot("plot 'z.L' us 1 title 'L(t)','z.L' us 2 title 'La(t)' with points,'z.L' us 3 title 'Lb(t)' with points");
  }
	
  return (Ltot*(Rmax-Rmin)+Rmin)*PtNorm;
}
